#ifdef HAVE_CONFIG_H
#include    "config.h"
#endif

#include	<stdlib.h>
#include	<errno.h>
#ifdef HAVE_STRING_H
#include	<string.h>
#endif

#ifdef HAVE_NETDB_H
#include	<netdb.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include	<sys/select.h>
#endif
#ifdef HAVE_NETDB_H
#include	<sys/socket.h>
#endif
#ifdef HAVE_NETDB_H
#include	<netinet/in.h>
#endif
#ifdef HAVE_W32API_BASETYPS_H
#include	<w32api/basetyps.h>
#endif
#ifdef _WINDOWS
#include	<winsock.h>
#endif
#include	<jni.h>

#include	"jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl.h"

void Throw( JNIEnv *env, const char *name, const char *msg);

static jfieldID	idFd;
static jfieldID	idFdNo;

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    socketInit
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_socketInit
  (JNIEnv *env, jclass cls)
{
	jclass clsFd;

	idFd = (*env)->GetFieldID( env, cls, "fd", "Ljava/io/FileDescriptor;" );
	clsFd = (*env)->FindClass( env, "java/io/FileDescriptor" );
	idFdNo = (*env)->GetFieldID( env, clsFd, "fd", "I" );
}

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    socketCreate
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_socketCreate
  (JNIEnv *env, jobject this, jint proto)
{
	int		fd;
	jobject	objFd;

	objFd = (*env)->GetObjectField( env, this, idFd );
	if ( objFd == NULL ) {
		Throw( env, "java/net/SocketException", "Socket closed" );
		return;
	}

	fd = socket( PF_INET, SOCK_RAW, proto );
	if ( fd < 0 ) {
		char msg[1024];
		sprintf( msg, "socket() error (%d)", errno );
		Throw( env, "java/net/SocketException", msg );
		return;
	}
	(*env)->SetIntField( env, objFd, idFdNo, fd );
}

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    socketClose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_socketClose
  (JNIEnv *env, jobject this)
{
	int		fd;
	jobject objFd;
	int		nRet;

	objFd = (*env)->GetObjectField( env, this, idFd );
	if ( objFd == NULL ) {
		Throw( env, "java/net/SocketException", "Socket closed" );
		return;
	}
	fd = (*env)->GetIntField( env, objFd, idFdNo );

#ifdef _WINDOWS
	nRet = closesocket(fd);
#else
	nRet = close(fd);
#endif
	if ( nRet < 0 ) {
		char msg[1024];
		sprintf( msg, "close error (%d)", errno );
		Throw( env, "java/net/SocketException", msg );
		return;
	}
	(*env)->SetIntField( env, objFd, idFdNo, -1 );
}

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    socketSetOption
 * Signature: (ILjava/lang/Object;)V
 */
JNIEXPORT void JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_socketSetOption
  (JNIEnv *env, jobject this, jint optId, jobject value)
{
	int		fd;
	int		nRet;
	jobject objFd;

	objFd = (*env)->GetObjectField( env, this, idFd );
	if ( objFd == NULL ) {
		Throw( env, "java/net/SocketException", "Socket closed" );
		return;
	}
	fd = (*env)->GetIntField( env, objFd, idFdNo );

	if ( optId == 1 ) {
		jclass clsVal = (*env)->FindClass( env, "java/lang/Integer" );
		jfieldID valID = (*env)->GetFieldID( env, clsVal, "value", "I" );
		int arg = (*env)->GetIntField( env, value, valID );
		nRet = setsockopt( fd, IPPROTO_IP, IP_TTL, (char *)&arg, sizeof(arg) );
/*
		if ( nRet < 0 ) {
			char msg[1024];
			sprintf( msg, "setsockopt error (%d)", errno );
			Throw( env, "java/net/SocketException", msg );
			return;
		}
*/
	}
}

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    socketGetOption
 * Signature: (I)[B
 */
JNIEXPORT jbyteArray JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_socketGetOption
  (JNIEnv *env, jobject this, jint optId)
{
}

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    bind
 * Signature: (Ljava/net/InetAddress;)V
 */
JNIEXPORT void JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_bind(
	JNIEnv *env, jobject this, jobject laddr)
{
	int		fd;
	jmethodID	method;
	struct hostent		*he;
	struct sockaddr_in	addr;
	jobject objFd;
	jclass clsAdr;
	jobject adrStr;
	const char *host;
	int nRet;

	objFd = (*env)->GetObjectField( env, this, idFd );
	if ( objFd == NULL ) {
		Throw( env, "java/net/SocketException", "Socket closed" );
		return;
	}
	fd = (*env)->GetIntField( env, objFd, idFdNo );

	clsAdr = (*env)->GetObjectClass( env, laddr );
	method = (*env)->GetMethodID( env, clsAdr, "getHostAddress",
								"()Ljava/lang/String;" );
	adrStr = (*env)->CallObjectMethod( env, laddr, method );

	host = (*env)->GetStringUTFChars( env, adrStr, NULL );
	he = gethostbyname( host );
	if ( he == NULL ) {
		char msg[1024];
		sprintf( msg, "unknown host [%s] (%d)", host, errno );
		Throw( env, "java/net/UnknownHostException", msg );
		return;
	}
	addr.sin_family = PF_INET;
	addr.sin_port = htons( 0 );
	addr.sin_addr = *((struct in_addr *)he->h_addr);
	memset( &addr.sin_zero, '\0', 8 );

	nRet = bind( fd, (struct sockaddr *)&addr, sizeof( addr ));
	if ( nRet < 0 ) {
		char msg[1024];
		sprintf( msg, "bind error [%s] (%d)", host, errno );
		Throw( env, "java/net/SocketException", msg );
		return;
	}
}

int setAddress( JNIEnv *env, jobject packet, struct sockaddr_in *addr ) {
	jclass clsPkt;
	jfieldID field;
	jobject jaddr;
	jclass clsAdr;
	jmethodID	method;
	jobject adrStr;
	const char *host;
	struct hostent		*he;

	clsPkt = (*env)->GetObjectClass( env, packet );
	field = (*env)->GetFieldID( env, clsPkt, "address",
								"Ljava/net/InetAddress;" );
	jaddr = (*env)->GetObjectField( env, packet, field );
	if ( jaddr == NULL ) {
		Throw( env, "java/lang/NullPointerException", "address is null" );
		return -1;
	}

	clsAdr = (*env)->GetObjectClass( env, jaddr );
	method = (*env)->GetMethodID( env, clsAdr, "getHostAddress",
								"()Ljava/lang/String;" );
	adrStr = (*env)->CallObjectMethod( env, jaddr, method );

	host = (*env)->GetStringUTFChars( env, adrStr, NULL );
	he = gethostbyname( host );
	if ( he == NULL ) {
		char msg[1024];
		sprintf( msg, "unknown host[%s] (%d)", host, errno );
		Throw( env, "java/net/UnknownHostException", msg );
		return -1;
	}
	addr->sin_family = PF_INET;
	addr->sin_port = htons( 0 );
	addr->sin_addr = *((struct in_addr *)he->h_addr);
	memset( &addr->sin_zero, '\0', 8 );
	return 0;
}

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    send
 * Signature: (Ljp/co/sds_corp/util/net/RawPacket;)V
 */
JNIEXPORT void JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_send
  (JNIEnv *env, jobject this, jobject packet)
{
	int		fd;
	jobject objFd;
	jclass clsPkt;
	jfieldID field;
	struct sockaddr_in  dest;
	int nLen, nOfs;
	int nRet;
	jobject jbuf;
	unsigned char	*buf;
	char msg[1024];

	objFd = (*env)->GetObjectField( env, this, idFd );
	if ( objFd == NULL ) {
		Throw( env, "java/net/SocketException", "Socket closed" );
		return;
	}
	fd = (*env)->GetIntField( env, objFd, idFdNo );

	nRet = setAddress( env, packet, &dest );
	if ( nRet < 0 ) {
		return;
	}

	clsPkt = (*env)->GetObjectClass( env, packet );
	field = (*env)->GetFieldID( env, clsPkt, "offset", "I" );
	nOfs = (*env)->GetIntField( env, packet, field );
	field = (*env)->GetFieldID( env, clsPkt, "length", "I" );
	nLen = (*env)->GetIntField( env, packet, field );

	field = (*env)->GetFieldID( env, clsPkt, "buf", "[B" );
	jbuf = (*env)->GetObjectField( env, packet, field );
	buf = (char *)malloc( (size_t)nLen );
	if ( buf == NULL ) {
		sprintf( msg, "malloc error (%d)", errno );
		Throw( env, "java/lang/NullPointerException", msg );
		return;
	}
	(*env)->GetByteArrayRegion( env, jbuf, nOfs, nLen, (jbyte *)buf );
/*
int i;
printf( "send: " );
for ( i = 0; i < nLen; i++ ) {
	printf( "%02x ", buf[i+nOfs] );
}
printf( "\n" );
*/
	nRet = sendto( fd, buf, nLen, 0,
					(struct sockaddr *)&dest, sizeof( struct sockaddr ));
	free( buf );
	if ( nRet < 0 ) {
		char msg[1024];
		sprintf( msg, "sendto error (%d)", errno );
		Throw( env, "java/net/SocketException", msg );
		return;
	}
}

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    peek
 * Signature: (Ljava/net/InetAddress;)V
 */
JNIEXPORT void JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_peek
  (JNIEnv *env, jobject this, jobject addr)
{
}

/*
 * Class:     jp_co_sds_0005fcorp_util_net_PlainRawSocketImpl
 * Method:    receive
 * Signature: (Ljp/co/sds_corp/util/net/RawPacket;)V
 */
JNIEXPORT void JNICALL Java_jp_co_sds_1corp_util_net_PlainRawSocketImpl_receive
  (JNIEnv *env, jobject this, jobject packet)
{
	int		fd;
	jfieldID field;
	jmethodID	method;
	char	*buf;
	struct sockaddr_in	dest;
	struct sockaddr_in	remote;
	int		nRemoteLen;
	int		nLen, nOfs;
	int		nRet;
	jobject objFd;
	jclass clsPkt;
	fd_set readfds;
	struct timeval tv;
	jclass clsHdr;
	jobject header;
	char b;
	char ihl;
	short total;
	jobject jbuf;
	int	len;
	time_t start;
	time_t now;
	char msg[1024];

	objFd = (*env)->GetObjectField( env, this, idFd );
	if ( objFd == NULL ) {
		Throw( env, "java/net/SocketException", "Socket closed" );
		return;
	}
	fd = (*env)->GetIntField( env, objFd, idFdNo );

	clsPkt = (*env)->GetObjectClass( env, packet );
	field = (*env)->GetFieldID( env, clsPkt, "offset", "I" );
	nOfs = (*env)->GetIntField( env, packet, field );
	field = (*env)->GetFieldID( env, clsPkt, "length", "I" );
	nLen = (*env)->GetIntField( env, packet, field );

	nRet = setAddress( env, packet, &dest );
	if ( nRet < 0 ) {
		return;
	}

	nRemoteLen = sizeof( remote );
	buf = (char *)malloc( nLen );
	if ( buf == NULL ) {
		sprintf( msg, "malloc error (%d)", errno );
		Throw( env, "java/lang/NullPointerException", msg );
		return;
	}
	start = time( NULL );
	for ( ;; ) {
		FD_ZERO( &readfds );
		FD_SET( fd, &readfds );

		tv.tv_sec = 5;
		tv.tv_usec = 0;

		nRet = select( fd + 1, &readfds, NULL, NULL, &tv );
		if ( nRet < 0 ) {
			free( buf );
			sprintf( msg, "select error (%d)", errno );
			Throw( env, "java/net/SocketException", msg );
			return;
		}
		if ( nRet == 0 ) {
			free( buf );
			Throw( env, "java/net/SocketTimeoutException", "select timeout" );
			return;
		}

		nRet = recvfrom( fd, buf, nLen, 0,
							(struct sockaddr *)&remote, &nRemoteLen );
		if ( nRet < 0 ) {
			free( buf );
			sprintf( msg, "receive error (%d)", errno );
			Throw( env, "java/net/SocketException", msg );
			return;
		}
/*{
unsigned char *da = (char *)&dest.sin_addr.s_addr;
unsigned char *ra = (char *)&remote.sin_addr.s_addr;
printf( "recv %d [%d.%d.%d.%d-%d.%d.%d.%d]\n", fd,
	*da, *(da+1), *(da+2), *(da+3), *ra, *(ra+1), *(ra+2), *(ra+3) );
}*/
		if ( memcmp( &dest.sin_addr.s_addr, &remote.sin_addr.s_addr, 4 ) == 0 ) {
			break;
		}
		now = time( NULL );
		if (( now - start ) > 5 ) {
			free( buf );
			sprintf( msg, "timeout error" );
			Throw( env, "java/net/SocketTimeoutException", msg );
			return;
		}
	}
/*
{int i;
char    msg[1024];
char    s[32];
sprintf( msg, "$$$ recv %d (%d:%d): ", fd, nLen, nRet );
for ( i = 0; i < nRet; i++ ) {
	sprintf( s, "%02x ", (unsigned char )buf[i] );
	strcat( msg, s );
}
printf( "%s\n", msg );}
*/
	clsHdr = (*env)->FindClass( env, "jp/co/sds_corp/util/net/IPHeader" );
	field = (*env)->GetFieldID( env, clsPkt, "header",
								"Ljp/co/sds_corp/util/net/IPHeader;" );
	header = (*env)->GetObjectField( env, packet, field );

	b = ( buf[0] & 0xf0 ) >> 4;
	field = (*env)->GetFieldID( env, clsHdr, "version", "B" );
	(*env)->SetByteField( env, header, field, b );
	ihl = buf[0] & 0xf;
	field = (*env)->GetFieldID( env, clsHdr, "ihl", "B" );
	(*env)->SetByteField( env, header, field, ihl );
	field = (*env)->GetFieldID( env, clsHdr, "ttl", "B" );
	(*env)->SetByteField( env, header, field, buf[8] );
	total = ntohs( *(jshort *)(buf + 2) );
	field = (*env)->GetFieldID( env, clsHdr, "totalLength", "I" );
	(*env)->SetShortField( env, header, field, total );

	field = (*env)->GetFieldID( env, clsHdr, "source", "[B" );
	jbuf = (*env)->GetObjectField( env, header, field );
	(*env)->SetByteArrayRegion( env, jbuf, 0, 4, (jbyte *)buf + 12 );
	field = (*env)->GetFieldID( env, clsHdr, "dest", "[B" );
	jbuf = (*env)->GetObjectField( env, header, field );
	(*env)->SetByteArrayRegion( env, jbuf, 0, 4, (jbyte *)buf + 16 );

	len = nRet - (ihl * 4);
	field = (*env)->GetFieldID( env, clsPkt, "length", "I" );
	jbuf = (*env)->GetObjectField( env, packet, field );
	(*env)->SetIntField( env, packet, field, len );

	field = (*env)->GetFieldID( env, clsPkt, "buf", "[B" );
	jbuf = (*env)->GetObjectField( env, packet, field );
	(*env)->SetByteArrayRegion( env, jbuf, 0, len, (jbyte *)&buf[ihl * 4] );
	free( buf );
}

void Throw( JNIEnv *env, const char *name, const char *msg) {
	jclass cls = (*env)->FindClass( env, name );

	if (cls != 0) {
		(*env)->ThrowNew( env, cls, msg );
	}
	(*env)->DeleteLocalRef( env, cls );
}
